;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "gptel-agent" "20260502.1809"
  "Agentic LLM use for gptel."
  '((emacs     "29.1")
    (compat    "30.1.0.0")
    (gptel     "0.9.9")
    (yaml      "1.2.0")
    (orderless "1.1"))
  :url "https://github.com/karthink/gptel-agent"
  :commit "aecaa940faf29107b4bc0a2bd621419c2d04b8d0"
  :revdesc "aecaa940faf2"
  :keywords '("comm"))
